/*WBL 10 Dec 2009 $Revision: 1.5 $
 * based on gp_cuda_kernel r1.111 */

/*WBL 29 Nov 2009 Skip r1.109 (using constant). For 295 GTX */


//#include <stdio.h>
//#include <stdlib.h>
#include <assert.h>
#include <cutil_inline.h>

//Fast integer multiplication
#define MUL(a, b) __umul24(a, b)
const int pStackMax = 15;
typedef unsigned char optype;
const optype OPNOP  = 0;

//Programming_Guide_2.0Beta2.pdf 4.2.2.3
//http://forums.nvidia.com/lofiversion/index.php?t64398.html
extern __shared__ int shared_array[];

//based on reduce6  ~/cuda/sdk/projects/harris/reduction/compute/projects/reduction/reduction_kernel.cu
__device__ inline void reduce_fit(const int thread_value, const int n, int *d_Output) {
  //Assumes n <= blockSize and both are powers of two between 1 and 256
  //Harris uses templates so that compiler evaluates many if ...
  //parallel reduction cf Harris' reduction_kernel.cu
/*
    This version adds multiple elements per thread sequentially.  This reduces the overall
    cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
    (Brent's Theorem optimization)
*/
  if (n == 1) {*d_Output = thread_value; return; /*avoid __syncthreads() etc*/ }
  int *sdata = (int*) shared_array; //Overlays Stack
  const int tid = threadIdx.x;
  const int lid = tid & (n-1);
  sdata[tid] = thread_value;
  __syncthreads();

  // do reduction in shared mem
  if (n >= 256) { if (lid <128) { sdata[tid] += sdata[tid + 128]; } __syncthreads();} //all threads must sync
  if (n >= 128) { if (lid < 64) { sdata[tid] += sdata[tid +  64]; } __syncthreads();} //all threads must sync

  if (lid < 32) {
    //__syncthreads() not needed as operate wholy within single warp 

    if (n >=  64) { sdata[tid] += sdata[tid + 32]; }
    if (n >=  32) { sdata[tid] += sdata[tid + 16]; }
    if (n >=  16) { sdata[tid] += sdata[tid +  8]; }
    if (n >=   8) { sdata[tid] += sdata[tid +  4]; }
    if (n >=   4) { sdata[tid] += sdata[tid +  2]; }
    if (n >=   2) { sdata[tid] += sdata[tid +  1]; }
  }
  // write result to global mem 
  if (lid == 0) *d_Output = sdata[tid];
}

//nb ok for sp1, sp2 and sp3 to _read_ from top
#define stack(sp) Stack[(sp)*blockDim.x+threadIdx.x]
#define SHAREDBYTESPERTHREAD ( pStackMax*sizeof(int) )

__device__ inline void runprog(unsigned char* const Pop,const int prog, const int u, const int LEN) {
#define push(x) {stack(SP) = x; SP++;}
  int* Stack = shared_array;

  stack(0) = 0;
  int SP = 0;
  for(int PC = 0; PC < LEN-1; PC++){
    const optype OPCODE = Pop[PC+(prog*LEN)];
    if(OPCODE==OPNOP) break;

    const optype data = (OPCODE + u) & 255;
    if(PC==0) {push(data);}
    else {
      const int sp1 = stack(SP-1);
      //jiggle SP across all of stack to test shared memory
      if(SP>=pStackMax || (OPCODE & 1)) {SP += (SP>1)? -2 : -1; }
      push(sp1+data);
    }
  }
  if(SP>1) stack(0) = stack(SP-1);
}


static __global__ void evalpopKernel(
    int *d_Output,
    unsigned char* Pop, 
    const int prog0, /*first program's number on host*/
    int N,
    int gen, int ntrain, const int npar2,
    const int LEN) {
    const int npar = 1<<npar2; //number of parallel fitness cases
    const int first_pid = (MUL(blockDim.x, blockIdx.x) + threadIdx.x)>>npar2;
    const int threadN   =  MUL(blockDim.x, gridDim.x) >> npar2;
    for(int prog = first_pid; prog < N; prog += threadN){
      int fitness = 0;
      for(int index = threadIdx.x & (npar-1); index < ntrain; index+=npar){
	int* Stack = shared_array;
	runprog(Pop,prog,index,LEN);
	fitness += stack(0);
      }//endfor testcases
      reduce_fit(fitness, npar, &d_Output[prog]);
    }
}


//Host-side interface
static void evalpopGPU(int *d_Output,
		       int grid_size, int block_size, unsigned char* Pop, const int prog0, int N, int gen, int ntests, int npar2, int LEN){
  assert(npar2>=0 && npar2<=8);
  //reduce_fit etc assumes block_size is a power of two
  assert(block_size==1 ||block_size==2 ||block_size==4  ||block_size==8||block_size==16||
	 block_size==32||block_size==64||block_size==128||block_size==256);
  assert((1<<npar2) <= block_size); //reduce_fit etc assumes npar evenly devides blocksize
  //assert((block_size & ((1<<npar2)-1))==0); //reduce_fit etc assumes npar evenly devides blocksize
  evalpopKernel<<<grid_size, block_size, block_size*SHAREDBYTESPERTHREAD>>>(d_Output, Pop, prog0, N, gen, ntests, npar2, LEN);
  cutilCheckMsg("evalpopKernel() execution failed.\n");
}


extern "C" void evalpop_gpu(int *d_Output, 
			    int grid_size, int block_size, unsigned char* Pop, const int prog0, int N, int gen, int ntests, int npar2, int LEN){
  evalpopGPU(d_Output, grid_size, block_size, Pop, prog0, N, gen, ntests, npar2, LEN);
}

